@@ -1,8 +1,6 @@
 #include "mktmpdir.h"
 #include "sha1.h"
 
-#define PAR_TEMP "PAR_TEMP"
-
 #ifdef O_BINARY
 #  define OPEN_O_BINARY O_BINARY
 #else
@@ -23,6 +21,20 @@
 #define Stat_t struct stat
 #endif
 
+static char PAR_MAGIC[] = "\nPAR.pm\n";
+#define magic_size 8
+
+/* "\0CACHE" or "\0CLEAN" */
+#define cache_marker_size 6
+
+/* size of a pack("N") number */
+#define FILE_offset_size 4
+
+#define cache_name_size 40
+
+#define CHECK(code, msg) if (!(code)) { par_die(msg); }
+
+
 static int isWritableDir(const char* val)
 {
     Stat_t statbuf;
@@ -51,24 +63,58 @@ static int isSafeDir(const char* val)
 
 void par_setup_libpath( const char * stmpdir )
 {
-    const char *val = NULL;
-    char *ld_path = LDLIBPTHNAME;
-    char *ld_path_env = NULL;
+    const char *ldlibpthname = stringify(LDLIBPTHNAME);
+    const char *val;
 
-    if ( (val = par_getenv(ld_path)) == NULL || strlen(val) == 0 ) {
-        par_setenv(ld_path, stmpdir);
+    if ( (val = par_getenv(ldlibpthname)) == NULL || strlen(val) == 0 ) {
+        par_setenv(ldlibpthname, stmpdir);
     }
-    else if ( !strstr(val, stmpdir) ) {
+    else {
         /* prepend stmpdir to (value of) environment variable */
-        ld_path_env = malloc( 
+       char *new_val = malloc( 
             strlen(stmpdir) + strlen(path_sep) + strlen(val) + 1);
         sprintf(
-            ld_path_env, "%s%s%s",
+            new_val, "%s%s%s",
             stmpdir, path_sep, val);
-        par_setenv(ld_path, ld_path_env);
+        par_setenv(ldlibpthname, new_val);
     }
 }
 
+static void *par_memrmem(void* haystack, size_t haystacklen, void* needle, size_t needlelen)
+{
+    char *hs = haystack;
+    char *p;
+    if (haystacklen < needlelen)
+        return NULL;
+    for (p = hs + haystacklen - needlelen; p >= hs; p--)
+        if (memcmp(p, needle, needlelen) == 0)
+            return p;
+    return NULL;
+}
+
+static off_t find_par_magic(int fd)
+{
+#define CHUNK_SIZE (64 * 1024)
+    char buf[CHUNK_SIZE + magic_size];
+    off_t pos;
+    int len;
+    char *p;
+    off_t file_size = lseek(fd, 0, 2);
+
+    for (pos = (file_size-1) - (file_size-1) % CHUNK_SIZE; 
+         pos >= 0; 
+         pos -= CHUNK_SIZE) {
+        CHECK(lseek(fd, pos, 0) != -1, "lseek failed");
+        len = read(fd, buf, CHUNK_SIZE + magic_size);
+        CHECK(len != -1, "read failed");
+        p = par_memrmem(buf, len, PAR_MAGIC, magic_size);
+        if (p)
+            return pos + (p - buf);
+    }
+    return -1;
+#undef CHUNK_SIZE
+}
+
 char *par_mktmpdir ( char **argv ) {
     int i;
     const char *tmpdir = NULL;
@@ -91,12 +137,9 @@ char *par_mktmpdir ( char **argv ) {
     char *progname = NULL, *username = NULL;
     char *stmpdir = NULL, *top_tmpdir = NULL;
     int f, j, k, stmp_len = 0;
-    char sha1[41];
-    SHA_INFO *sha_info;
-    unsigned char buf[32768];
-    unsigned char sha_data[20];
+    char sha1[cache_name_size + 1];
 
-    if ( (val = par_getenv(PAR_TEMP)) && strlen(val) ) {
+    if ( (val = par_getenv("PAR_TEMP")) && strlen(val) ) {
         par_setup_libpath(val);
         return strdup(val);
     }
@@ -141,12 +184,12 @@ char *par_mktmpdir ( char **argv ) {
 #ifdef WIN32
     /* Try the windows temp directory */
     if ( tmpdir == NULL && (val = par_getenv("WinDir")) && strlen(val) ) {
-        char* buf = malloc(strlen(val) + 5 + 1);
-        sprintf(buf, "%s\\temp", val);
-        if (isWritableDir(buf)) {
-            tmpdir = buf;
+        char* p = malloc(strlen(val) + 5 + 1);
+        sprintf(p, "%s\\temp", val);
+        if (isWritableDir(p)) {
+            tmpdir = p;
         } else {
-            free(buf);
+            free(p);
         }
     }
 #endif
@@ -208,11 +251,12 @@ char *par_mktmpdir ( char **argv ) {
 #else
 #define STREQ(a,b) (strcmp(a,b) == 0)
 #endif
+        const char *parl_exe = stringify(PARL_EXE);
 	int prog_len = strlen(progname);
-	int parl_len = strlen(PARL_EXE);
+	int parl_len = strlen(parl_exe);
 
 	if (prog_len >= parl_len
-	    && STREQ(progname + prog_len - parl_len, PARL_EXE)
+	    && STREQ(progname + prog_len - parl_len, parl_exe)
 	    && (prog_len == parl_len || progname[prog_len - parl_len - 1] == dir_sep[0])
 	    && argv[1]
 	    && strlen(argv[1]) >= 4
@@ -221,47 +265,39 @@ char *par_mktmpdir ( char **argv ) {
 #undef STREQ
     }
 
+    int use_cache = 0;
     if ( !par_env_clean() && (f = open( progname, O_RDONLY | OPEN_O_BINARY ))) {
-        lseek(f, -18, 2);
-        read(f, buf, 6);
-        if(buf[0] == 0 && buf[1] == 'C' && buf[2] == 'A' && buf[3] == 'C' && buf[4] == 'H' && buf[5] == 'E') {
-            /* pre-computed cache_name in this file */
-            /* "$TEMP/par-$USER/cache-$cache_name" */
-            lseek(f, -58, 2);
-            read(f, buf, 41);
-            sprintf(
-                stmpdir,
-                "%s%scache-%s%s",
-                top_tmpdir, dir_sep, buf, subdirbuf_suffix
-            );
-        }
-        else {
-            /* "$TEMP/par-$USER/cache-$SHA1" */
-	    lseek(f, 0, 0);
-            sha_info = sha_init();
-            while( ( j = read( f, buf, sizeof( buf ) ) ) > 0 )
-            {
-                sha_update( sha_info, buf, j );
+        off_t pos = find_par_magic(f);
+        char buf[cache_marker_size];
+
+        if (pos >= 0) {                 
+            /* back up over pack(N) number and "\0CACHE" (or "\0CLEAN") */
+            pos -= FILE_offset_size + cache_marker_size;                  
+            lseek(f, pos, 0); 
+            CHECK(read(f, buf, cache_marker_size) == cache_marker_size, "short read");
+            if (memcmp(buf, "\0CACHE", cache_marker_size) == 0) {
+                use_cache = 1;
+                /* back up over pre-computed cache_name */
+                pos -= cache_name_size;
+                lseek(f, pos, 0);
+                CHECK(read(f, sha1, cache_name_size) == cache_name_size, "short read");
+                sha1[cache_name_size] = '\0';
             }
-            close( f );
-            sha_final( sha_data, sha_info );
-            for( k = 0; k < 20; k++ )
-            {
-                sprintf( sha1+k*2, "%02x", sha_data[k] );
+            else if (memcmp(buf, "\0CLEAN", cache_marker_size) == 0) {
+                use_cache = 0;
             }
-            sha1[40] = '\0';
-            sprintf(
-                stmpdir,
-                "%s%scache-%s%s",
-                top_tmpdir, dir_sep, sha1, subdirbuf_suffix
-            );
         }
     }
+    if (use_cache) {
+        /* "$TEMP/par-$USER/cache-$SHA1" */
+        sprintf(
+            stmpdir,
+            "%s%scache-%s%s",
+            top_tmpdir, dir_sep, sha1, subdirbuf_suffix
+        );
+    }
     else {
-        int i = 0;
-
         /* "$TEMP/par-$USER/temp-$PID" */
-
         par_setenv("PAR_CLEAN", "1");
         sprintf(
             stmpdir,
@@ -274,7 +310,7 @@ char *par_mktmpdir ( char **argv ) {
            "$TEMP/par-$USER/temp-$PID-$i". This will guard against cases where
            a prior invocation crashed leaving garbage in a temp directory that
            might interfere. */
-
+        int i = 0;
         while (my_mkdir(stmpdir, 0700) == -1 && errno == EEXIST) {
             sprintf(
                 stmpdir,
@@ -287,7 +323,7 @@ char *par_mktmpdir ( char **argv ) {
     free(top_tmpdir);
 
     /* set dynamic loading path */
-    par_setenv(PAR_TEMP, stmpdir);
+    par_setenv("PAR_TEMP", stmpdir);
 
     par_setup_libpath( stmpdir );
 
@@ -379,9 +415,10 @@ static void par_rmtmpdir ( char *stmpdir ) {
 #endif
 
 void par_cleanup (char *stmpdir) {
-    char *dirname = par_dirname(stmpdir);
-    char *basename = par_basename(dirname);
+    char *dirname, *basename;
     if ( par_env_clean() && stmpdir != NULL && strlen(stmpdir)) {
+        dirname = par_dirname(stmpdir);
+        basename = par_basename(dirname);
         if ( strstr(basename, "par-") == basename ) {
             par_rmtmpdir(stmpdir);
             /* Don't try to remove dirname because this will introduce a race
